#ifndef __LUNAR_KADEMLIASTORAGE_H__
#define __LUNAR_KADEMLIASTORAGE_H__

#include <utility>

#include <string>
#include <list>
//test
#include <iostream>

#include <boost/multi_index_container.hpp>
#include <boost/multi_index/hashed_index.hpp>
#include <boost/multi_index/ordered_index.hpp>
#include <boost/multi_index/member.hpp>
#include <boost/multi_index/identity.hpp>
#include <boost/cstdint.hpp>

const boost::uint32_t KADEMLIA_STORAGE_TIMEOUT = 60;

typedef std::pair<std::string, std::string> STORE_ITEM;

/**
 * This class is implement a map struture for local use
 * for now, this class stays in Kademlia namespace,
 * maybe later made into more general form and moved into mechanisms.
 * 
 * current implementation use multimap as underlying DS.
 * hash_map is not applicable, it does not support key->multiple value
 * map is not applicable either, more than 1 index are needed 
 *
 * 1. tread-saft
 * 2. two index on the DS -> time/key
 * 3. one thread to delete outdated key/value every 15 min. 
 */

enum STORAGETYPE
  {
    ST_NOTYPE = 0,
    ST_SINGLE,
    ST_MULT,
    ST_NUM
  };

enum STORAGE_OP_RESULT
  {
    ST_RT_OK,
    ST_RT_FAILED_NOREASON,
    ST_RT_BAD_PARAMETER,
    ST_RT_FAILED_DUPKEY,
    ST_RT_TOOMANYREQUESTS,
    ST_RT_DONE_REQUESTS_PENDING,
    ST_RT_NO_PENDING_DONE_REQUESTS,
  };

const boost::uint32_t ST_DEFAULT_TIMEOUT = 12 * 60 * 60;

class MKademliaStorage
{
public:
    bool Initialize(STORAGETYPE st);

    ~MKademliaStorage()
    {
    }

    STORAGE_OP_RESULT Put(const std::string& key, const STORE_ITEM& item, time_t timeout = ST_DEFAULT_TIMEOUT);
    STORAGE_OP_RESULT Put(const std::string& key, const std::list<STORE_ITEM>& items, time_t timeout = ST_DEFAULT_TIMEOUT);

    STORAGE_OP_RESULT Get(const std::string& key, std::list<STORE_ITEM>& items);

    STORAGE_OP_RESULT Del(const std::string& key, const std::string& version);
    STORAGE_OP_RESULT Del(const std::string& key, std::list<std::string>& version);
    STORAGE_OP_RESULT Del(time_t t);

    void Dump();
	void Refresh();

private:
    //    friend class MSingleton<MKademliaStorage>;
    //    MKademliaStorage(){}
    //
    STORAGE_OP_RESULT Put_Special(const std::string& key, const STORE_ITEM& item, time_t timeout = ST_DEFAULT_TIMEOUT);


    void ProcessTimeout(void* param);

    struct Item
    {
        Item(const std::string& k, time_t t, const std::string& v, const std::string& vv) : _key(k), _timeout(t), _value(v), _version(vv)
        {
        }

        std::string _key;
        time_t _timeout;
        std::string _value;
        std::string _version;

        bool operator<(const Item& i) const
        {
            return _timeout < i._timeout;
        }
    };

    struct KEY
    {
    };
    struct TIME
    {
    };
    typedef boost::multi_index_container<Item, boost::multi_index::indexed_by<boost::multi_index::hashed_non_unique<boost::multi_index::tag<KEY>, boost::multi_index::member<Item, std::string, &Item::_key> >, boost::multi_index::ordered_non_unique<boost::multi_index::tag<TIME>, boost::multi_index::member<Item, time_t, &Item::_timeout> > > > Store;

    struct MTimeComp
    {
		bool operator()(boost::uint32_t n, const MKademliaStorage::Item& i) const
        {
			return n < (boost::uint32_t)i._timeout;
        }
		bool operator()(const MKademliaStorage::Item& i, boost::uint32_t n) const
        {
			return (boost::uint32_t)i._timeout < n;
        }
    };

    struct Mod
    {
        Mod(const std::string& v, time_t to) : _v(v), _to(to)
        {
        }

        void operator()(MKademliaStorage::Item& item)
        {
            //std::cerr << item._key << "<><>" << item._value << "<><>" << item._version << std::endl;

            item._value   = _v;
            item._timeout = _to;
        }

        std::string _v;
        time_t _to;
    };

    //      typedef employee_set::index<name>::type employee_set_by_name;

    typedef Store::index<KEY>::type::iterator StoreIter;
    typedef Store::index<KEY>::type::const_iterator StoreConstIter;
    typedef std::pair<StoreIter, StoreIter> StoreRange;

    typedef Store::index<TIME>::type::iterator StoreTimeIter;
    typedef Store::index<TIME>::type::const_iterator StoreTimeConstIter;
    typedef std::pair<StoreTimeIter, StoreTimeIter> StoreTimeRange;

    Store _st;
};

#endif
